from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Literal

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.signals import setting_changed
from django.utils.module_loading import import_string


def maybe_import_string(val: Any | None):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_string(val)
    return val


SettingKey = Literal["CORE", "FRONTEND", "CODEGEN", "STORAGE", "AUDIT", "UI", "PDF"]


class AlliancePlatformSettingsBase:
    _has_loaded = False
    # This is a dict of settings, type will be one of the TypedDicts defined in the settings module. Typing as Any
    # to avoid errors like `Incompatible types in assignment` - can't strictly type this as this common module is copied
    # into each package and doesn't know anything about them
    _user_settings: Any = {}

    _key: SettingKey
    _defaults: dict[str, Any]
    _import_strings: list[str]

    def __init__(
        self, key: SettingKey, defaults: dict | None = None, import_strings: list[str] | None = None
    ):
        self._key = key
        self._defaults = defaults or {}
        self._import_strings = import_strings or []
        self._path_strings = [k for k, v in self.__class__.__annotations__.items() if v is Path]

        setting_changed.connect(self._reload_api_settings)

    def _reload_api_settings(self, *args, **kwargs):
        setting = kwargs["setting"]
        if setting == "ALLIANCE_PLATFORM":
            self.reload()

    def _load_user_settings(self):
        self._has_loaded = True
        try:
            platform_settings = settings.ALLIANCE_PLATFORM
        except AttributeError:
            raise ImproperlyConfigured(
                "Missing ALLIANCE_PLATFORM settings dict - check that you have not imported alliance_platform.frontend.settings before defining your ALLIANCE_PLATFORM settings dict"
            )
        try:
            self._user_settings = platform_settings[self._key]
        except KeyError:
            # Don't throw an error if not set so things work with packages that only have optional settings
            self._user_settings = {}

    @lru_cache()
    def __getattr__(self, key):
        if not self._has_loaded:
            self._load_user_settings()
        try:
            value = self._user_settings.get(key) if key in self._user_settings else self._defaults[key]
        except KeyError:
            raise ImproperlyConfigured(
                f"Missing required setting '{key}' in ALLIANCE_PLATFORM[\"{self._key}\"] settings"
            )
        if isinstance(value, LazySetting):
            value = value()
        if key in self._import_strings:
            try:
                value = maybe_import_string(value)
            except ImportError as e:
                msg = f"Could not import '{value}' for ALLIANCE_PLATFORM[\"{self._key}\"] setting '{key}'. {e.__class__.__name__}: {e}."
                raise ImportError(msg)
        if key in self._path_strings and isinstance(value, str):
            value = Path(value)
        return value

    def reload(self):
        self.__getattr__.cache_clear()
        self._load_user_settings()

    def check_settings(self):
        """Can be implemented to check settings are valid when app is ready"""
        pass


class LazySetting:
    def __init__(self, getter):
        self.getter = getter

    def __call__(self):
        return self.getter()



from typing import overload

from alliance_platform.core.settings import ap_core_settings
from django.apps import AppConfig
from django.db.models import Model


def default_resolve_perm_name(
    app_config: AppConfig,
    model: Model | type[Model] | None,
    action: str,
    is_global: bool,
) -> str:
    """Default implementation of :func:`~alliance_platform.core.auth.resolve_perm_name`.

    .. note::

        Don't call this directly; use :func:`~alliance_platform.core.auth.resolve_perm_name` which will call the function
        set in :data:`~alliance_platform.core.settings.AlliancePlatformCoreSettingsType.RESOLVE_PERM_NAME` (which defaults
        to ``default_resolve_perm_name``).

    Returns strings in the following format:

        With a model:     ``{app_config_label}.{model_name}_{action}``, e.g. ``"myapp.mymodel_list"``
        Without a model:  ``{app_config_label}.{action}``, e.g. ``"myapp.management"``

    We follow a different convention to django (model_action rather than action_model) so that permission
    names sort lexicographically.

    The default implementation makes no use of the ``is_global`` parameter.

    Args:
        app_config: The app config the permission is for. All permissions are scoped to an app. If ``model``, you can pass
            ``model._meta.app_config`` here..
        model: The model to use in the permission name. Can be ``None`` if permission is not specific to a model.
        action: The action to perform. For example, common ones include ``"create"``, ``"update"``, ``"detail"``, ``"list"``, ``"delete"``,
            but it can be anything.
        is_global: Whether the permission is global (``True``) or per-object (``False``). The default implementation does
            not use this parameter, but it is required for compatability with :data:`~alliance_platform.core.settings.AlliancePlatformCoreSettingsType.RESOLVE_PERM_NAME`.
    """
    if model:
        return f"{app_config.label}.{model._meta.model_name}_{action}"
    return f"{app_config.label}.{action}"


@overload
def resolve_perm_name(entity: Model | type[Model], action: str, is_global: bool) -> str:
    """Resolve a permission name for a model and action."""
    ...


@overload
def resolve_perm_name(entity: AppConfig, action: str, is_global: bool) -> str:
    """Resolve a permission name for an app and action"""
    ...


def resolve_perm_name(
    entity: Model | type[Model] | AppConfig,
    action: str,
    is_global: bool,
) -> str:
    """Resolve a permission name for an app or model, and an action.

    This calls the function set in :data:`~alliance_platform.core.settings.AlliancePlatformCoreSettingsType.RESOLVE_PERM_NAME`,
    which defaults to :func:`~alliance_platform.core.auth.default_resolve_perm_name`.

    This method can be used when writing generic code that needs to generate a default permission name according to
    a convention. For example, on generic CRUD views.

    Usage for model permissions:

    .. code-block:: pycon

        >>> resolve_perm_name(MyModel, "list", True)
        "myapp.mymodel_list"
        >>> record = MyModel.objects.get(pk=5)
        >>> resolve_perm_name(record, "update", False)
        "myapp.mymodel_update"

    For permissions that are not specific to a model, pass an :class:`~django.apps.AppConfig` instead of a model:

    .. code-block:: pycon

        >>> from django.apps import apps
        >>> resolve_perm_name(apps.get_app_config("MyApp"), "dashboard", True)
        "myapp.dashboard"

    .. note::

        The default implementation makes no use of the ``is_global`` parameter, but custom implementations may so it
        is required.

    Args:
        entity: This will either by an :class:`~django.apps.AppConfig`, or a :class:`~django.db.models.Model` class or instance.
            All permissions are scoped to an app; if a model is passed the config will be extracted from the model.
        action: The action to perform. For example, common ones for a model include ``"create"``, ``"update"``, ``"detail"``,
            ``"list"``, ``"delete"``, but it can be anything.
        is_global: Whether the permission is global (``True``) or per-object (``False``).
    """
    if isinstance(entity, Model) or (isinstance(entity, type) and issubclass(entity, Model)):
        app_config = entity._meta.app_config
        model = entity
    elif isinstance(entity, AppConfig):
        app_config = entity
        model = None
    else:
        raise ValueError("First argument must be a model or an app config.")

    return ap_core_settings.RESOLVE_PERM_NAME(app_config, model, action, is_global)



from alliance_platform.core.settings import AlliancePlatformCoreSettingsType
from django.conf import settings
from django.test import override_settings
from typing_extensions import Unpack


class override_ap_core_settings(override_settings):
    def __init__(self, **kwargs: Unpack[AlliancePlatformCoreSettingsType]):
        CORE = settings.ALLIANCE_PLATFORM.get("CORE", {})
        super().__init__(
            ALLIANCE_PLATFORM={
                **settings.ALLIANCE_PLATFORM,
                "CORE": {
                    **CORE,
                    **kwargs,
                },
            }
        )



import datetime
import decimal
from typing import Any

from django.utils import formats
from django.utils import timezone


def default_display_for_value(value: Any, empty_value_display: str | None = ""):
    """Return a string representation of a value for display to user

    Based on :meth:`django.contrib.admin.utils.display_for_value`

    Any project specific customisations should be added to :meth:`xenopus_frog_app.display.display_for_value`
    so that it's easier to pull in upstream changes.

    Args:
        value: The value to generate representation for
        empty_value_display: If ``value`` is ``None`` this will be returned instead
    """
    if value is None:
        return empty_value_display
    if isinstance(value, datetime.datetime):
        t = timezone.template_localtime(value)  # type: ignore[attr-defined] # template_localtime not in django-stubs
        return formats.localize(t)
    if isinstance(value, (datetime.date, datetime.time)):
        return formats.localize(value)
    # `bool` inherits from `int` so handle it before number_format below
    if isinstance(value, bool):
        return str(value)
    if isinstance(value, (int, decimal.Decimal, float)):
        return formats.number_format(value)
    if isinstance(value, (list, tuple)):
        return ", ".join(str(v) for v in value)
    return str(value)



from alliance_platform.core.auth import resolve_perm_name
from django.apps import apps
from django.test import TestCase
from test_alliance_platform_core.models import PermTestModelA
from test_alliance_platform_core.models import PermTestModelB

from .test_utils import override_ap_core_settings


class TestAuth(TestCase):
    def test_default_resolve_perm_name(self):
        self.assertEqual(
            resolve_perm_name(PermTestModelA, "list", True),
            "test_alliance_platform_core.permtestmodela_list",
        )

        self.assertEqual(
            resolve_perm_name(PermTestModelB(), "update", False),
            "test_alliance_platform_core.permtestmodelb_update",
        )

        self.assertEqual(
            resolve_perm_name(apps.get_app_config("test_alliance_platform_core"), "management", True),
            "test_alliance_platform_core.management",
        )

    def test_customised_resolve_perm_name(self):
        for setting_value in [resolve_perm_name_for_testing, "tests.test_auth.resolve_perm_name_for_testing"]:
            with self.subTest(setting_value=setting_value):
                with override_ap_core_settings(RESOLVE_PERM_NAME=setting_value):
                    self.assertEqual(
                        resolve_perm_name(PermTestModelA, "list", True),
                        "test_alliance_platform_core|permtestmodela|list|True",
                    )

                    self.assertEqual(
                        resolve_perm_name(PermTestModelB, "update", False),
                        "test_alliance_platform_core|permtestmodelb|update|False",
                    )

                    self.assertEqual(
                        resolve_perm_name(
                            apps.get_app_config("test_alliance_platform_core"), "management", True
                        ),
                        "test_alliance_platform_core|management|True",
                    )


def resolve_perm_name_for_testing(app_config, model, action, is_global):
    if model:
        return f"{app_config.label}|{model._meta.model_name}|{action}|{is_global}"
    return f"{app_config.label}|{action}|{is_global}"



from pathlib import Path
from typing import Callable
from typing import TypedDict

from alliance_platform.base_settings import AlliancePlatformSettingsBase
from alliance_platform.base_settings import LazySetting
from django.apps import AppConfig
from django.db.models import Model

RESOLVE_PERM_NAME_TYPE = Callable[[AppConfig, Model | type[Model] | None, str, bool], str]


class AlliancePlatformCoreSettingsType(TypedDict, total=False):
    """Settings for the core package of the Alliance Platform

    These can be set in the Django settings file under the ``ALLIANCE_PLATFORM`` key:

    .. code-block:: python

        ALLIANCE_PLATFORM = {
            "CORE": {
                "PROJECT_DIR": PROJECT_DIR,
            }
        }
    """

    #: The root directory of the project
    PROJECT_DIR: Path | str
    #: A directory used for caching. This is used by various packages. If not set, defaults to ``PROJECT_DIR / '.alliance-platform'``
    CACHE_DIR: Path | str | None
    RESOLVE_PERM_NAME: str | RESOLVE_PERM_NAME_TYPE
    """
    A function used to resolve a permission name for a model and action. You can set this to function, or an import
    path to a function. The function should have the signature:
    
    .. code-block:: python
    
        def custom_resolve_perm_name(
            # When `model` is passed, `app_config` will be set to `model._meta.app_config`
            app_config: AppConfig,
            # The model class or instance
            model: Model | type[Model] | None,
            # The name of the action. This can be any string, but common ones for CRUD actions
            # are "create", "update", "detail", "list", "delete".
            action: str,
            # Whether the permission is global (``True``) or per-object (``False``). The
            # default implementation does not make use of this parameter, but a custom
            # implementation may.
            is_global: bool
        ) -> str: ...
        
    This is used by the :func:`~alliance_platform.core.auth.resolve_perm_name` function, and is used throughout
    alliance_platform when a generic default permission is needed for a model or app. You can see
    :func:`~alliance_platform.core.auth.default_resolve_perm_name` for an example, and the default used
    when none is provided.
    """


class AlliancePlatformCoreSettings(AlliancePlatformSettingsBase):
    PROJECT_DIR: Path
    CACHE_DIR: Path
    RESOLVE_PERM_NAME: RESOLVE_PERM_NAME_TYPE


DEFAULTS = dict(
    CACHE_DIR=LazySetting(lambda: ap_core_settings.PROJECT_DIR / ".alliance-platform"),  # type: ignore[has-type]
    RESOLVE_PERM_NAME="alliance_platform.core.auth.default_resolve_perm_name",
)

ap_core_settings = AlliancePlatformCoreSettings(
    "CORE",
    defaults=DEFAULTS,
    import_strings=["RESOLVE_PERM_NAME"],
)



# Taken verbatim from distutils source in Python 3.11
# https://github.com/python/cpython/blob/v3.11.2/Lib/distutils/util.py#L308
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))
